use std::{net::SocketAddr, time::Duration};

use lanspread_db::db::Game;
use lanspread_proto::{Message as _, Request, Response};
use lanspread_utils::maybe_addr;
use s2n_quic::{client::Connect, provider::limits::Limits, Client as QuicClient};
use tokio::sync::mpsc::{UnboundedReceiver, UnboundedSender};

static CERT_PEM: &str = include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/../../cert.pem"));

#[derive(Debug)]
pub enum ClientEvent {
    ListGames(Vec<Game>),
}

#[derive(Debug)]
pub enum ClientCommand {
    ListGames,
    GetGame(u64),
    ServerAddr(SocketAddr),
}

/// # Errors
pub async fn run(
    mut rx_control: UnboundedReceiver<ClientCommand>,
    tx_event: UnboundedSender<ClientEvent>,
) -> eyre::Result<()> {
    // blocking wait for remote address
    log::debug!("waiting for server address");
    let server_addr = loop {
        if let Some(ClientCommand::ServerAddr(addr)) = rx_control.recv().await {
            log::info!("got server address: {addr}");
            break addr;
        }
    };

    loop {
        let limits = Limits::default()
            .with_max_handshake_duration(Duration::from_secs(3))?
            .with_max_idle_timeout(Duration::from_secs(1))?;

        let client = QuicClient::builder()
            .with_tls(CERT_PEM)?
            .with_io("0.0.0.0:0")?
            .with_limits(limits)?
            .start()?;

        let conn = Connect::new(server_addr).with_server_name("localhost");
        let mut conn = match client.connect(conn).await {
            Ok(conn) => conn,
            Err(e) => {
                log::error!("failed to connect to server: {e}");
                tokio::time::sleep(Duration::from_secs(3)).await;
                continue;
            }
        };

        conn.keep_alive(true)?;

        log::info!(
            "connected: (server: {}) (client: {})",
            maybe_addr!(conn.remote_addr()),
            maybe_addr!(conn.local_addr())
        );

        // tx
        while let Some(cmd) = rx_control.recv().await {
            let request = match cmd {
                ClientCommand::ListGames => Request::ListGames,
                ClientCommand::GetGame(id) => Request::GetGame { id },
                ClientCommand::ServerAddr(_) => Request::Invalid(
                    [].into(),
                    "invalid control message (ServerAddr), should not happen".into(),
                ),
            };

            let data = request.encode();
            log::error!("encoded data: {}", String::from_utf8_lossy(&data));

            let stream = match conn.open_bidirectional_stream().await {
                Ok(stream) => stream,
                Err(e) => {
                    log::error!("failed to open stream: {e}");
                    break;
                }
            };

            let (mut rx, mut tx) = stream.split();

            if let Err(e) = tx.send(data).await {
                log::error!("failed to send request to server {:?}", e);
            }

            if let Ok(Some(data)) = rx.receive().await {
                log::trace!("server response (raw): {}", String::from_utf8_lossy(&data));

                let response = Response::decode(&data);
                log::trace!(
                    "server response (decoded): {}",
                    String::from_utf8_lossy(&data)
                );
                match response {
                    Response::Games(games) => {
                        for game in &games {
                            log::debug!("{game:?}");
                        }

                        if let Err(e) = tx_event.send(ClientEvent::ListGames(games)) {
                            log::error!("failed to send ClientEvent::ListGames to client {e:?}");
                        }
                    }
                    Response::Game(game) => log::debug!("game received: {game:?}"),
                    Response::GameNotFound(id) => log::debug!("game not found {id}"),
                    Response::InvalidRequest(request_bytes, err) => log::error!(
                        "server says our request was invalid (error: {}): {}",
                        err,
                        String::from_utf8_lossy(&request_bytes)
                    ),
                    Response::EncodingError(err) => {
                        log::error!("server encoding error: {err}");
                    }
                    Response::DecodingError(data, err) => {
                        log::error!(
                            "response decoding error: {} (data: {})",
                            err,
                            String::from_utf8_lossy(&data)
                        );
                    }
                }

                if let Err(err) = tx.close().await {
                    log::error!("failed to close stream: {err}");
                }
            }
        }
    }

    // log::info!("server closed connection");
    // Ok(())
}

// #[derive(Debug, Parser)]
// struct Cli {
//     /// Server IP address.
//     #[clap(long, default_value = "127.0.0.1")]
//     ip: IpAddr,
//     /// Server port.
//     #[clap(long, default_value = "13337")]
//     port: u16,
// }

// #[tokio::main]
// async fn main() -> eyre::Result<()> {
//     let cli = Cli::parse();

//     let (tx_control, rx_control) = tokio::sync::mpsc::unbounded_channel::<ControlMessage>();

//     // Spawn client in a separate task
//     let client_handle = tokio::spawn(async move {
//         let remote_addr = SocketAddr::from((cli.ip, cli.port));
//         Client::run(remote_addr, rx_control).await
//     });

//     // Handle stdin commands in the main task
//     let mut stdin = BufReader::new(tokio::io::stdin());
//     let mut line = String::new();

//     loop {
//         line.clear();
//         if stdin.read_line(&mut line).await? == 0 {
//             break; // EOF reached
//         }

//         // Trim whitespace and handle commands
//         match line.trim() {
//             "list" => {
//                 tx_control.send(ControlMessage::ListGames)?;
//             }
//             cmd if cmd.starts_with("get ") => {
//                 if let Ok(id) = cmd[4..].trim().parse::<u64>() {
//                     tx_control.send(ControlMessage::GetGame(id))?;
//                 } else {
//                     println!("Invalid game ID");
//                 }
//             }
//             "quit" | "exit" => break,
//             "" => continue,
//             _ => println!("Unknown command. Available commands: list, get <id>, quit"),
//         }
//     }

//     client_handle.await??;
//     Ok(())
// }
