use std::{
    net::{IpAddr, SocketAddr},
    path::{Path, PathBuf},
    sync::Arc,
};

use clap::Parser;
use lanspread_db::db::GameDB;
use lanspread_mdns::{
    DaemonEvent,
    MdnsAdvertiser,
    LANSPREAD_INSTANCE_NAME,
    LANSPREAD_SERVICE_TYPE,
};
use lanspread_proto::{Message as _, Request, Response};
use lanspread_utils::maybe_addr;
use s2n_quic::Server as QuicServer;
use testing::generate_test_db;
use tokio::{io::AsyncWriteExt, sync::Mutex};
use tracing_subscriber::EnvFilter;

mod testing;

static KEY_PEM: &str = include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/../../key.pem"));
static CERT_PEM: &str = include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/../../cert.pem"));

pub(crate) struct Server {
    db_path: PathBuf,
}

#[derive(Clone, Debug)]
struct ServerCtx {
    handler: RequestHandler,
}

#[derive(Clone, Debug)]
struct ConnectionCtx {
    server_ctx: Arc<ServerCtx>,
    remote_addr: String,
}

impl Server {
    fn new<S: Into<PathBuf>>(db_path: S) -> Self {
        Server {
            db_path: db_path.into(),
        }
    }

    async fn run(&mut self, addr: SocketAddr) -> eyre::Result<()> {
        let mut server = QuicServer::builder()
            .with_tls((CERT_PEM, KEY_PEM))?
            .with_io(addr)?
            .start()?;

        let server_ctx = Arc::new(ServerCtx {
            handler: RequestHandler::new(&self.db_path)?,
        });

        while let Some(mut connection) = server.accept().await {
            let conn_ctx = Arc::new(ConnectionCtx {
                server_ctx: server_ctx.clone(),
                remote_addr: maybe_addr!(connection.remote_addr()),
            });
            // spawn a new task for the connection
            tokio::spawn(async move {
                tracing::info!("{} connected", conn_ctx.remote_addr);

                while let Ok(Some(stream)) = connection.accept_bidirectional_stream().await {
                    let (mut rx, mut tx) = stream.split();

                    let conn_ctx = conn_ctx.clone();
                    // spawn a new task for the stream
                    tokio::spawn(async move {
                        tracing::trace!("{} stream opened", conn_ctx.remote_addr);

                        // handle streams
                        while let Ok(Some(data)) = rx.receive().await {
                            tracing::trace!(
                                "{} client request (raw): {}",
                                conn_ctx.remote_addr,
                                String::from_utf8_lossy(&data)
                            );

                            let request = Request::decode(&data);
                            tracing::debug!(
                                "{} client request (decoded): {:?}",
                                conn_ctx.remote_addr,
                                request
                            );
                            let response =
                                conn_ctx.server_ctx.handler.handle_request(request).await;
                            tracing::trace!(
                                "{} server response: {:?}",
                                conn_ctx.remote_addr,
                                response
                            );
                            let raw_response = response.encode();
                            tracing::trace!(
                                "{} server response (raw): {}",
                                conn_ctx.remote_addr,
                                String::from_utf8_lossy(&raw_response)
                            );

                            // write response back to client
                            if let Err(e) = tx.write_all(&raw_response).await {
                                tracing::error!(?e);
                            }

                            // close the stream
                            if let Err(e) = tx.close().await {
                                tracing::error!(?e);
                            }
                        }
                    });
                }
            });
        }

        Ok(())
    }
}

#[derive(Clone, Debug)]
struct RequestHandler {
    db: Arc<Mutex<GameDB>>,
}

impl RequestHandler {
    fn new(db_path: &Path) -> eyre::Result<Self> {
        let db = GameDB::load_from_file(db_path)?;
        Ok(RequestHandler {
            db: Arc::new(Mutex::new(db)),
        })
    }

    async fn handle_request(&self, request: Request) -> Response {
        match request {
            Request::ListGames => {
                let db = self.db.lock().await;
                Response::Games(db.all_games().into_iter().cloned().collect())
            }
            Request::GetGame { id } => {
                let db = self.db.lock().await;
                match db.get_game_by_id(id) {
                    Some(game) => Response::Game(game.clone()),
                    None => Response::GameNotFound(id),
                }
            }
            Request::Invalid(data, err_msg) => {
                tracing::error!(
                    "got invalid request from client (error: {}): {}",
                    err_msg,
                    String::from_utf8_lossy(&data)
                );
                Response::InvalidRequest(data, err_msg)
            }
        }
    }
}

const GAME_DB_PATH: &str = "/home/pfs/shm/game.db";

#[derive(Debug, Parser)]
struct Cli {
    /// The IP address to bind to.
    #[clap(long, default_value = "127.0.0.1")]
    ip: IpAddr,
    /// The listen port.
    #[clap(long, default_value = "13337")]
    port: u16,
}

#[tokio::main]
async fn main() -> eyre::Result<()> {
    tracing_subscriber::fmt()
        .with_env_filter(EnvFilter::from_default_env())
        .init();

    let cli = Cli::parse();

    generate_test_db(GAME_DB_PATH);

    let mdns = MdnsAdvertiser::new(
        LANSPREAD_SERVICE_TYPE,
        LANSPREAD_INSTANCE_NAME,
        (cli.ip, cli.port).into(),
    )?;

    tokio::spawn(async move {
        while let Ok(event) = mdns.monitor.recv() {
            tracing::info!("mDNS: {:?}", &event);
            if let DaemonEvent::Error(e) = event {
                tracing::info!("Failed: {e}");
                break;
            }
        }
    });

    tracing::info!("Server listening on {}:{}", cli.ip, cli.port);

    let mut server = Server::new(GAME_DB_PATH);
    server.run(SocketAddr::from((cli.ip, cli.port))).await
}
