use js_sys::Function;
use lanspread_db::db::Game;
use leptos::*;
use serde::{Deserialize, Serialize};
use wasm_bindgen::prelude::*;
use web_sys::console::log_1;

#[wasm_bindgen]
extern "C" {
    #[wasm_bindgen(js_namespace = ["window", "__TAURI__", "core"])]
    async fn invoke(cmd: &str, args: JsValue) -> JsValue;

    #[wasm_bindgen(js_namespace = ["window", "__TAURI__", "event"])]
    async fn listen(event: &str, handler: &Function) -> JsValue;
}

#[derive(Serialize, Deserialize)]
struct RunGameArgs {
    id: u64,
}

#[derive(Serialize, Deserialize)]
enum CommandArgs {
    RunGame(RunGameArgs),
}

#[component]
pub fn App() -> impl IntoView {
    let (game_items, set_game_items) = create_signal(Vec::<Game>::new());

    // Something that listens to the global app state and updates the game_items signal
    create_effect(move |_| {
        spawn_local(async move {
            // Convert the callback to a JS function that can be passed to listen()
            let callback = Closure::wrap(Box::new(move |event: JsValue| {
                log_1(&JsValue::from_str("Received games-list-updated event"));

                // Access the payload property of the event
                let payload = js_sys::Reflect::get(&event, &JsValue::from_str("payload")).unwrap();

                // Parse the event payload
                match serde_wasm_bindgen::from_value::<Vec<Game>>(payload) {
                    Ok(games) => {
                        for game in &games {
                            log_1(&JsValue::from_str(format!("game: {game}").as_str()));
                        }
                        set_game_items.update(|items| *items = games);
                    }
                    Err(e) => log_1(&JsValue::from_str(
                        format!("Error parsing games-list-updated event: {e}").as_str(),
                    )),
                }
            }) as Box<dyn FnMut(JsValue)>);

            // Register the event listener
            listen("games-list-updated", callback.as_ref().unchecked_ref()).await;
            callback.forget(); // Prevent callback from being dropped
        });
    });

    // Call list_games on component mount
    create_effect(move |_| {
        spawn_local(async move {
            let args = serde_wasm_bindgen::to_value(&()).unwrap();
            invoke("request_games", args).await;
        });
    });

    let run_game = move |id: u64| {
        log_1(&JsValue::from_str(format!("id={id}").as_str()));
        spawn_local(async move {
            let args = serde_wasm_bindgen::to_value(&RunGameArgs { id }).unwrap();
            let result = invoke("run_game_backend", args).await.as_string().unwrap();
            log_1(&JsValue::from_str(format!("id={result}").as_str()));
        });
    };

    view! {
        <main class="container">
            <h1 class="align-center">{"SoftLAN Launcher"}</h1>

            <div class="main-header">
                {"HEADER"}
            </div>

            <div class="grid-container">
                {move || game_items.get().into_iter()
                    .map(|item|
                        view! {
                            <div class="item" on:click=move |_| run_game(item.id)>
                                <img src="https://via.placeholder.com/200x150" alt="Item Image" />
                                <div class="item-name">{ &item.name }</div>
                                <div class="description">
                                    <span class="desc-text">{ &item.description }</span>
                                    <span class="size-text">{ &item.size.to_string() }</span>
                                </div>
                                <div class="play-button">{"Play"}</div>
                            </div>
                        }
                    ).collect::<Vec<_>>()
                }
            </div>
        </main>
    }
}
